from __future__ import annotations #  for 3.9 compatability
import abc
from typing import TYPE_CHECKING, IO, Any, Self
import textwrap
import inspect


import numpy.ma
import numpy as np
from archnemesis.helpers.io_helper import OutWidth

from .ModelParameterEntry import ModelParameterEntry
from .ModelParameter import ModelParameter



if TYPE_CHECKING:
    # NOTE: This is just here to make 'flake8' play nice with the type hints
    # the problem is that importing Variables_0 or ForwardModel_0 creates a circular import
    # this actually means that I should possibly redesign how those work to avoid circular imports
    # but that is outside the scope of what I want to accomplish here
    from archnemesis.Variables_0 import Variables_0
    from archnemesis.ForwardModel_0 import ForwardModel_0
    
    nx = 'number of elements in state vector'
    m = 'an undetermined number, but probably less than "nx"'
    mx = 'synonym for nx'
    mparam = 'the number of parameters a model has'
    NCONV = 'number of spectral bins'
    NGEOM = 'number of geometries'
    NX = 'number of elements in state vector'
    NDEGREE = 'number of degrees in a polynomial'
    NWINDOWS = 'number of spectral windows'

import logging
_lgr = logging.getLogger(__name__)
#_lgr.setLevel(logging.DEBUG)
_lgr.setLevel(logging.INFO)




class ModelBase(abc.ABC):
    """
        Abstract base class of all parameterised models used by ArchNemesis. This class should be subclassed further for models of a particular component.
    """
    
    id : int = None # All "*ModelBase" classes that are not meant to be used should have an id of 'None'
    
    @classmethod
    def to_string(cls):
        desc_wrapper_first = textwrap.TextWrapper(
            width = OutWidth.get(),
            expand_tabs=True,
            tabsize=2,
            replace_whitespace=True,
            initial_indent='- description: ',
            subsequent_indent=' '*15
        )
        desc_wrapper_rest = textwrap.TextWrapper(
            width = OutWidth.get(),
            expand_tabs=True,
            tabsize=2,
            replace_whitespace=True,
            initial_indent=' '*15,
            subsequent_indent=' '*15
        )
        
        ds : list[str,...] = textwrap.dedent(cls.__doc__.strip('\n') if cls.__doc__ is not None else 'DESCRIPTION NOT FOUND').split('\n\n')
        
        docstr = textwrap.indent('\n'.join((
            desc_wrapper_first.fill(x) if i==0 else desc_wrapper_rest.fill(x) for i,x in enumerate(ds)
        )), '|', lambda x: True)
        return '\n'.join((
            f'{cls.__name__}:',
            f'|- id : {cls.id}',
            f'|- parent classes: {", ".join((x.__name__ for x in cls.__bases__))}',
            docstr,
        ))
    
    def __init__(
            self, 
            state_vector_start : int, 
            #   The index of the first entry of the model parameters in the state vector
            
            n_state_vector_entries : int,
            #   The number of model parameters that are stored in the state vector
        ):
        """
            Initialise an instance of the model.
            
            ## ARGUMENTS ##
                
                state_vector_start : int
                    The index of the first entry of the model parameters in the state vector
                
                n_state_vector_entries : int
                    The number of model parameters that are stored in the state vector
            
            ## RETURNS ##
                
                None
        """
        
        # Store where the model parameters are positioned within the state vector
        self.state_vector_start = state_vector_start
        self.n_state_vector_entries = n_state_vector_entries
        self.state_vector_slice = slice(state_vector_start, state_vector_start+n_state_vector_entries)
        
        # default parameters, the default will make a fake "all_parameters" parameter
        # that contains everything.
        self.parameters = (
            ModelParameter('all_parameters', slice(None), f'default parameter that is all of the parameters for model id {self.id}', 'UNDEFINED'),
        )
        
        _lgr.debug(f'{self.id=} {self.state_vector_start=} {self.n_state_vector_entries=}')
        return
    
    
    def info(self, log_flag_state_vector = None, apriori_state_vector=None, posterior_state_vector=None)->str:
        param_apriori_values = None
        param_posterior_values = None
        if log_flag_state_vector is not None and apriori_state_vector is not None:
            param_apriori_values = self.get_parameter_values_from_state_vector(apriori_state_vector, log_flag_state_vector)
            _lgr.debug(f'{param_apriori_values=}')
        if log_flag_state_vector is not None and posterior_state_vector is not None:
            param_posterior_values = self.get_parameter_values_from_state_vector(posterior_state_vector, log_flag_state_vector)
        
        return self.__str__(param_apriori_values, param_posterior_values)
    
    
    def __str__(self, param_apriori_values = None, param_posterior_values = None) -> str:
        s = self.to_string()
        #attrs = tuple(filter(lambda x: not x.startswith('_') and x != "parameters", dir(self)))
        attrs = inspect.getmembers(self, lambda x: not inspect.ismethod(x))
        attrs = tuple(filter(lambda x: not x[0].startswith('_') and x[0] not in ("parameters", "id"), attrs))
        attrs_str = '\n'.join(
            (
                f'|- {name} : {attr}' for name, attr in attrs
            )
        )
        param_desc_wrapper_first = textwrap.TextWrapper(
            width = OutWidth.get(),
            expand_tabs=True,
            tabsize=2,
            replace_whitespace=False,
            initial_indent='- description: ',
            subsequent_indent=' '*15
        )
        param_desc_wrapper_rest = textwrap.TextWrapper(
            width = OutWidth.get(),
            expand_tabs=True,
            tabsize=2,
            replace_whitespace=False,
            initial_indent=' '*15,
            subsequent_indent=' '*15
        )
        
        
        
        params_strs = [
            '\n'.join((
                f'|  |- {p.name} :',
                f'|  |  |- slice : {p.slice}',
                f'|  |  |- unit : {p.unit}',
                textwrap.indent(
                    '\n'.join((
                        param_desc_wrapper_first.fill(x) if i==0 else param_desc_wrapper_rest(x) for i,x in enumerate(p.description.strip().split('\n'))
                    )),
                    '|  |  |',
                    lambda x: True
                ),
            )) for p in self.parameters
        ]
        
        for i in range(len(params_strs)):
            if param_apriori_values is not None:
                params_strs[i] += f'\n|  |  |- apriori value : {param_apriori_values[i]}'
            if param_posterior_values is not None:
                params_strs[i] += f'\n|  |  |- posterior value : {param_posterior_values[i]}'
        
        
        return '\n'.join((
            s,
            attrs_str,
            '|- Parameters:',
            '\n'.join(params_strs),
        ))
        
    
    def get_state_vector_slice(
            self, 
            state_vector : np.ndarray[['nx'], float|int],
            #   Array that we want to pull from. Will normally be the apriori or posterior state vector.
            
        ) -> np.ndarray[['mparam'],float|int]:
        """
            Gets the slice of a `state_vector`-like object (i.e. same shape and ordering as state vector)
            that holds only the parameters for the model.
        """
        return state_vector[self.state_vector_slice]
    
    
    def get_parameter_entries_from_state_vector(
            self,
            state_vector : np.ndarray[['nx'],float],
            #   Array that we want to pull from. Will normally be the apriori or posterior state vector.
            
        ) -> tuple[float | np.ndarray[[int],float],...]:
        """
            Returns a tuple of copies of the entries (raw entries, not unlogged etc.) of parameters (defined via `self.parameters`).
            Parameters with a single entry will be a 'float', parameters that consist of multiple entries will be a numpy array.
            
            ## RETURNS ##
            
                entries : tuple[float | np.ndarray[[int],float],...]
                    Entries retrieved from the state vector, no processing is done to the entries.
        """
        sv_slice = self.get_state_vector_slice(state_vector)
        return tuple(np.array(sv_slice[p.slice]) if len(sv_slice[p.slice]) > 1 else sv_slice[p.slice][0] for p in self.parameters)
    
    
    def get_parameter_values_from_state_vector(
            self,
            state_vector : np.ndarray[['nx'],float],
            #   Array that we want to pull from. Will normally be the apriori or posterior state vector.
            
            state_vector_log : np.ndarray[['nx'],int],
            #   Array of boolean flags indicating if the value stored in `state_vector` is the exponential logarithm
            #   of the 'real' value.
            
        ) -> tuple[float | np.ndarray[[int],float],...]:
        """
            Returns a tuple of copies of the values (entry will be unlogged etc. if required) of parameters (defined via `self.parameters`).
            Parameters with a single entry will be a 'float', parameters that consist of multiple entries will be a numpy array.
            
            ## RETURNS ##
            
                values : tuple[float | np.ndarray[[int],float],...]
                    Values retrieved from the state vector, unlogged as required depending upon the `state_vector_log` array entries.
        """
        sv_slice = self.get_state_vector_slice(state_vector)
        sv_slice_exp = np.exp(sv_slice)
        log_slice = (self.get_state_vector_slice(state_vector_log) > 0)
        values = np.where(log_slice, sv_slice_exp, sv_slice)
        return tuple(np.array(values[p.slice]) if len(values[p.slice]) > 1 else values[p.slice][0] for p in self.parameters)
    
    
    def get_value_from_state_vector(
            self,
            state_vector : np.ndarray[['nx'],float],
            #   Array that we want to pull from. Will normally be the apriori or posterior state vector.
            
            state_vector_log : np.ndarray[['nx'],int],
            #   Array of boolean flags indicating if the value stored in `state_vector` is the exponential logarithm
            #   of the 'real' value.
            
            sub_slice : slice = slice(None),
            #   A sub-slice that is applied after the state vector is sliced the first time to only contain elements
            #   associated with the model. For example this can be set to 'slice(0,1)' to only get the first element
            #   of the state vector associated with the model, useful when splitting up the "whole model" state vector
            #   to get each individual parameter of the model.
            
        ) -> np.ndarray[['m'],float]:
        """
            Returns the value (entry will be unlogged if required) of elements of a (sub-slice of a) state vector associated with the model.
            
            ## RETURNS ##
                
                value : np.ndarray[['m'],float]
                    Array of 'real' (i.e. unlogged where applicable) values of (a sub-slice of) the parameters of the model.
        """
        
        a_val = state_vector[self.state_vector_slice][sub_slice]
        a_exp = np.exp(a_val)
        a_log_flag = state_vector_log[self.state_vector_slice][sub_slice] != 0
        
        return np.where(a_log_flag, a_exp, a_val)
    
    def set_value_to_state_vector(
            self,
            value : np.ndarray[['m'],float],
            #   Array of values we want to put into the `state_vector` array
            
            state_vector : np.ndarray[['nx'],float],
            #   Array that we want to push to. Will normally be the apriori or posterior state vector.
            
            state_vector_log : np.ndarray[['nx'],int],
            #   Array of boolean flags indicating if the value stored in `state_vector` is the exponential logarithm
            #   of the 'real' value.
            
            sub_slice : slice = slice(None),
            #   A sub-slice that is applied after the state vector is sliced the first time to only contain elements
            #   associated with the model. For example this can be set to 'slice(0,1)' to only get the first element
            #   of the state vector associated with the model, useful when splitting up the "whole model" state vector
            #   to get each individual parameter of the model.
            
        ) -> None:
        """
            Sets the value (will be logged if required) of elements of a (sub-slice of a) state vector associated with the model.
            
            ## RETURNS ##
                
                None
        """
        
        value_log = np.log(value)
        log_flag = state_vector_log[self.state_vector_slice][sub_slice] != 0
        
        state_vector[self.state_vector_slice][sub_slice] = np.where(log_flag, value_log, value)
        return
    
    
    def get_parameters_from_state_vector(
            self,
            apriori_state_vector : np.ndarray[['nx'],float],
            #   The complete apriori state vector with 'nx' entries
            
            posterior_state_vector : np.ndarray[['nx'],float],
            #   The complete posterior state vector with 'nx' entries
            
            state_vector_log : np.ndarray[['nx'],int],
            #   Array of 'log' flags for each entry in either state vector (they share these flags)
            #   if the flag is non-zero the value stored in both state vectors is the exponential
            #   logarithm of the 'real' value.
                
            state_vector_fix : np.ndarray[['nx'],int],
            #   Array of the 'fix' flags for each entry in either state vector (they share these flags)
            #   if the flag is non-zero, the value stored in both state vectors is not retrieved, and
            #   therefore should be the same in each of the apriori/posterior state vectors.
            
        ) -> dict[str, ModelParameterEntry]:
        """
            Retrieve parameters from state vector as a dictionary of name : value pairs
            
            ## ARGUMENTS ##
                
                apriori_state_vector : np.ndarray[['nx'],float]
                    The complete apriori state vector with 'nx' entries
                
                posterior_state_vector : np.ndarray[['nx'],float]
                    The complete posterior state vector with 'nx' entries
                
                state_vector_log : np.ndarray[['nx'],int]
                    Array of 'log' flags for each entry in either state vector (they share these flags)
                    if the flag is non-zero the value stored in both state vectors is the exponential
                    logarithm of the 'real' value.
                
                state_vector_fix : np.ndarray[['nx'],int]
                    Array of the 'fix' flags for each entry in either state vector (they share these flags)
                    if the flag is non-zero, the value stored in both state vectors is not retrieved, and
                    therefore should be the same in each of the apriori/posterior state vectors.
            
            ## RETURNS ##
                
                parameters : dict[str, ModelParameterEntry]
                    A dictionary that maps parameter names to the "ModelParameterEntry" associated with that parameter.
        """
        parameters = dict()
        
        assert self.state_vector_slice.step is None or self.state_vector_slice.step == 1, "A step larger than 1 is not supported when slicing a state vector"
        
        for p in self.parameters:
            
            assert p.slice.step is None or p.slice.step == 1, "A step larger than 1 is not supported when sub-slicing a state vector"
            
            apriori_value = self.get_value_from_state_vector(
                apriori_state_vector, 
                state_vector_log, 
                p.slice
            )
            posterior_value = self.get_value_from_state_vector(
                posterior_state_vector, 
                state_vector_log, 
                p.slice
            )
            
            fix_flag = self.get_state_vector_slice(state_vector_fix)[p.slice] != 0
            
            p_start, p_stop, p_step = p.slice.indices(self.n_state_vector_entries)
            parameters[p.name] = ModelParameterEntry(
                self.id,
                p.name,
                slice(self.state_vector_slice.start + p_start, self.state_vector_slice.start + p_stop),
                fix_flag,
                apriori_value,
                posterior_value
            )
        return parameters
    
    
    
    def set_parameters_to_state_vector(
            self,
            parameter_entries : dict[str,ModelParameterEntry],
            #   A dictionary that maps parameter names to the "ModelParameterEntry" associated with that parameter.
            
            state_vector_log : np.ndarray[['nx'],int],
            #   Array of 'log' flags for each entry in either state vector (they share these flags)
            #   if the flag is non-zero the value stored in both state vectors is the exponential
            #   logarithm of the 'real' value.
            
            apriori_state_vector : np.ndarray[['nx'],float],
            #   The complete apriori state vector with 'nx' entries
            
            posterior_state_vector : np.ndarray[['nx'],float] | None = None,
            #   The complete posterior state vector with 'nx' entries
            
            
            
        ) -> dict[str, ModelParameterEntry]:
        """
            Retrieve parameters from state vector as a dictionary of name : value pairs
            
            ## ARGUMENTS ##
                
                model_parameter_entries : dict[str, ModelParameterEntry]
                    A dictionary that maps parameter names to the "ModelParameterEntry" associated with that parameter.
                    
                apriori_state_vector : np.ndarray[['nx'],float]
                    The complete apriori state vector with 'nx' entries
                
                posterior_state_vector : np.ndarray[['nx'],float]
                    The complete posterior state vector with 'nx' entries
                
                state_vector_log : np.ndarray[['nx'],int]
                    Array of 'log' flags for each entry in either state vector (they share these flags)
                    if the flag is non-zero the value stored in both state vectors is the exponential
                    logarithm of the 'real' value.
                
                state_vector_fix : np.ndarray[['nx'],int]
                    Array of the 'fix' flags for each entry in either state vector (they share these flags)
                    if the flag is non-zero, the value stored in both state vectors is not retrieved, and
                    therefore should be the same in each of the apriori/posterior state vectors.
            
            ## RETURNS ##
        """
        
        assert self.state_vector_slice.step is None or self.state_vector_slice.step == 1, "A step larger than 1 is not supported when slicing a state vector"
        
        for p in self.parameters:
            
            self.set_value_to_state_vector(
                parameter_entries[p.name].apriori_value,
                apriori_state_vector,
                state_vector_log,
                p.slice
            )
            
            if posterior_state_vector is not None:
                self.set_value_to_state_vector(
                    parameter_entries[p.name].posterior_value,
                    posterior_state_vector,
                    state_vector_log,
                    p.slice
                )
        return
    
    ## Abstract methods below this line, subclasses must implement all of these methods ##
    
    @classmethod
    @abc.abstractmethod
    def is_varident_valid(
            cls,
            varident : np.ndarray[[3],int],
            #   "Variable Identifier" from a *.apr file. Consists of 3 integers. Exact interpretation depends on the model
            #   subclass.
            
        ) -> bool:
        """
            Accepts a varident from a *.apr file, returns True if the varident is compatible with the model, False otherwise.
            Should be overwritten by a subclass
            
            ## ARGUMENTS ##
            
                varident : np.ndarra[[3],int]
                    "Variable Identifier" from a *.apr file. Consists of 3 integers. Exact interpretation depends on the model
                    subclass.
            
            ## RETURNS ##
            
                flag : bool
                    True if varident is compatible with the model, False otherwise.
        """
        ...
    
    
    @classmethod
    @abc.abstractmethod
    def from_apr_to_state_vector(
            cls,
            variables : "Variables_0", 
            #   An instance of the archnemesis.Variables_0.Variables_0 class that is reading the *.apr file
            
            f : IO, 
            #   The open file descriptor of the *.apr file
            
            varident : np.ndarray[[3],int], 
            #   Should be the correct slice of the original (which should be a reference to the sub-array)
            
            varparam : np.ndarray[["mparam"],float], 
            #   Should be the correct slice of the original (which should be a reference to the sub-array)
            
            ix : int, 
            #   The next free entry in the state vector
            
            lx : np.ndarray[["mx"],int], 
            #   state vector flags denoting if the value in the state vector is a logarithm of the 'real' value. 
            #   Should be a reference to the original
            
            x0 : np.ndarray[["mx"],float], 
            #   state vector, holds values to be retrieved. Should be a reference to the original
            
            sx : np.ndarray[["mx","mx"],float], 
            #   Covariance matrix for the state vector. Should be a reference to the original
            
            inum : np.ndarray[["mx"],int], 
            #   state vector flags denoting if the gradient is to be numerically calculated (1) 
            #   or analytically calculated (0) for the state vector entry. Should be a reference to the original
            
            npro : int, 
            #   Number of altitude levels defined for the atmosphere component
            
            ngas : int,
            #   Number of gas volume mixing ratio profiles defined for the reference atmosphere
            
            ndust : int,
            #   Number of aerosol species density profiles define for the reference atmosphere

            nlocations : int, 
            #   Number of locations defined for the atmosphere component
            
            runname : str, 
            #   Name of the *.apr file, without extension.
            
            sxminfac : float, 
            #   Minimum factor to bother calculating covariance matrix entries between current model's 
            #   parameters and other model's parameters.
            
        ) -> Self:
        """
            Constructs a model from its entry in a *.apr file. Should be overwritten by a subclass
            
            ## ARGUMENTS ##
            
                variables : Variables_0
                    The "Variables_0" instance that is reading the *.apr file
                
                f : IO
                    An open file descriptor for the *.apr file.
                
                varident : np.ndarray[[3],int]
                    "Variable Identifier" from a *.apr file. Consists of 3 integers. Exact interpretation depends on the model
                    subclass.
                
                varparam : np.ndarray[["mparam"], float]
                    "Variable Parameters" from a *.apr file. Holds "extra parameters" for the model. Exact interpretation depends on the model
                    subclass. NOTE: this is a holdover from the FORTRAN code, the better way to give extra data to the model is to store it on the
                    model instance itself.
                
                ix : int
                    The index of the next free entry in the state vector
                
                lx : np.ndarray[["mx"],int]
                    State vector flags denoting if the value in the state vector is a logarithm of the 'real' value. 
                    Should be a reference to the original
                
                x0 : np.ndarray[["mx"],float]
                    The actual state vector, holds values to be retrieved. Should be a reference to the original
                
                sx : np.ndarray[["mx","mx"],float]
                    Covariance matrix for the state vector. Should be a reference to the original
                
                inum : np.ndarray[["mx"],int]
                    state vector flags denoting if the gradient is to be numerically calulated (1) 
                    or analytically calculated (0) for the state vector entry. Should be a reference to the original
                
                npro : int
                    Number of altitude levels defined for the atmosphere component of the retrieval setup.
                
                ngas : int,
                    Number of gas volume mixing ratio profiles defined for the reference atmosphere
                
                ndust : int,
                    Number of aerosol species density profiles define for the reference atmosphere
                
                n_locations : int
                    Number of locations defined for the atmosphere component of the retrieval setup.
                
                runname : str
                    Name of the *.apr file, without extension. For example '/path/to/neptune.apr' has 'neptune'
                    as `runname`
                
                sxminfac : float
                    Minimum factor to bother calculating covariance matrix entries between current 
                    model's parameters and another model's parameters.
            
            
            ## RETURNS ##
            
                instance : Self
                    A constructed instance of the model class that has parameters set from information in the *.apr file
        """
        
        ...
    

    @classmethod
    @abc.abstractmethod
    def from_bookmark(
            cls,
            variables : "Variables_0", 
            #   An instance of the archnemesis.Variables_0 class
            
            varident : np.ndarray[[3],int], 
            #   Should be the correct slice of the original (which should be a reference to the sub-array)
            
            varparam : np.ndarray[["mparam"],float], 
            #   Should be the correct slice of the original (which should be a reference to the sub-array)
            
            ix : int, 
            #   The next free entry in the state vector
            
            npro : int, 
            #   Number of altitude levels defined for the atmosphere component
            
            ngas : int,
            #   Number of gas volume mixing ratio profiles defined for the reference atmosphere
            
            ndust : int,
            #   Number of aerosol species density profiles define for the reference atmosphere

            nlocations : int, 
            #   Number of locations defined for the atmosphere component
        ) -> Self:
        """
            Constructs a model from its entry in a *.apr file. Should be overwritten by a subclass
            
            ## ARGUMENTS ##
            
                variables : Variables_0
                    The "Variables_0" instance that enables acccess to `variables.classify_model_type_from_varident`
                
                varident : np.ndarray[[3],int]
                    "Variable Identifier" from a *.apr file. Consists of 3 integers. Exact interpretation depends on the model
                    subclass.
                
                varparam : np.ndarray[["mparam"], float]
                    "Variable Parameters" from a *.apr file. Holds "extra parameters" for the model. Exact interpretation depends on the model
                    subclass. NOTE: this is a holdover from the FORTRAN code, the better way to give extra data to the model is to store it on the
                    model instance itself.
                
                ix : int
                    The index of the next free entry in the state vector
                
                npro : int
                    Number of altitude levels defined for the atmosphere component of the retrieval setup.
                
                ngas : int,
                    Number of gas volume mixing ratio profiles defined for the reference atmosphere
                
                ndust : int,
                    Number of aerosol species density profiles define for the reference atmosphere
                
                n_locations : int
                    Number of locations defined for the atmosphere component of the retrieval setup.
            
            ## RETURNS ##
            
                instance : Self
                    A constructed instance of the model class that has parameters set from information in the *.apr file
        """
        
        ...

    
    
    @classmethod
    @abc.abstractmethod
    def calculate(
            cls, 
            *args, 
            **kwargs
        ) -> Any:
        """
            This class method should perform the lowest-level calculation for the model. Note that it is a class
            method (so we can easily call it from other models if need be) so you must pass any instance attributes
            as arguments.
            
            NOTE: Models are so varied in here that I cannot make any specific interface at this level of abstraction.
        """
        ...
    
    @abc.abstractmethod
    def calculate_from_subprofretg(
            self,
            forward_model : "ForwardModel_0",
            #   The ForwardModel_0 instance that is calling this function. We need this so we can alter components of the forward model
            #   inside this function.
            
            ix : int,
            #   The index of the state vector that corresponds to the start of the model's parameters
            
            ipar : int,
            #   An integer that encodes which part of the atmospheric component of the forward model this model should alter. Only
            #   used for some Atmospheric models.
            
            ivar : int,
            #   The model index, the order in which the models were instantiated. NOTE: this is a vestige from the
            #   FORTRAN version of the code, we don't really need to know this as we should be: 1) storing any model-specific
            #   values on the model instance itself; 2) passing any model-specific data from the outside directly
            #   instead of having the model instance look it up from a big array. However, the code for each model
            #   was recently ported from a more FORTRAN-like implementation so this is still required by some of them
            #   for now.
            
            xmap : np.ndarray,
            #   Functional derivatives of the state vector w.r.t Atmospheric profiles at each Atmosphere location.
            #   The array is sized as [nx,NVMR+2+NDUST,NP,NLOCATIONS]:
            #    
            #       nx - number of state vector entries.
            #    
            #       NVMR - number of gas volume mixing ratio profiles in the Atmosphere component of the forward model.
            #    
            #       NDUST - number of aerosol profiles in the Atmosphere component of the forward model.
            #    
            #       NP - number of points in an atmospheric profile, all profiles in an Atmosphere component of the forward model 
            #               should have the same number of points.
            #    
            #       NLOCATIONS - number of locations defined in the Atmosphere component of the forward model.
            #
            #   The size of the 1st dimension (NVMR+2+NDUST) is like that because it packs in 4 different atmospheric profile
            #   types: gas volume mixing ratios (NVMR), aerosol densities (NDUST), fractional cloud cover (1), para H2 fraction (1).
            
        ) -> None:
        """
            Updated values of components based upon values of model parameters in the state vector. Called from ForwardModel_0::subprofretg.
            
            ## ARGUMENTS ##
            
                forward_model : ForwardModel_0
                    The ForwardModel_0 instance that is calling this function. We need this so we can alter components of the forward model
                    inside this function.
                
                ix : int
                    The index of the state vector that corresponds to the start of the model's parameters
                    
                ipar : int
                    An integer that encodes which part of the atmospheric component of the forward model this model should alter. Only
                    used for some Atmospheric models.
                
                ivar : int
                    The model index, the order in which the models were instantiated. NOTE: this is a vestige from the
                    FORTRAN version of the code, we don't really need to know this as we should be: 1) storing any model-specific
                    values on the model instance itself; 2) passing any model-specific data from the outside directly
                    instead of having the model instance look it up from a big array. However, the code for each model
                    was recently ported from a more FORTRAN-like implementation so this is still required by some of them
                    for now.
                
                xmap : np.ndarray[[nx,NVMR+2+NDUST,NP,NLOCATIONS],float]
                    Functional derivatives of the state vector w.r.t Atmospheric profiles at each Atmosphere location.
                    The array is sized as:
                        
                        nx - number of state vector entries.
                        
                        NVMR - number of gas volume mixing ratio profiles in the Atmosphere component of the forward model.
                        
                        NDUST - number of aerosol profiles in the Atmosphere component of the forward model.
                        
                        NP - number of points in an atmospheric profile, all profiles in an Atmosphere component of the forward model 
                             should have the same number of points.
                        
                        NLOCATIONS - number of locations defined in the Atmosphere component of the forward model.
                    
                    The size of the 1st dimension (NVMR+2+NDUST) is like that because it packs in 4 different atmospheric profile
                    types: gas volume mixing ratios (NVMR), aerosol densities (NDUST), fractional cloud cover (1), para H2 fraction (1).
                    
                
            ## RETURNS ##
            
                None
        """
        ...

    @abc.abstractmethod
    def patch_from_subprofretg(
            self,
            forward_model : "ForwardModel_0",
            #   The ForwardModel_0 instance that is calling this function. We need this so we can alter components of the forward model
            #   inside this function.
            
            ix : int,
            #   The index of the state vector that corresponds to the start of the model's parameters
            
            ipar : int,
            #   An integer that encodes which part of the atmospheric component of the forward model this model should alter. Only
            #   used for some Atmospheric models.
            
            ivar : int,
            #   The model index, the order in which the models were instantiated. NOTE: this is a vestige from the
            #   FORTRAN version of the code, we don't really need to know this as we should be: 1) storing any model-specific
            #   values on the model instance itself; 2) passing any model-specific data from the outside directly
            #   instead of having the model instance look it up from a big array. However, the code for each model
            #   was recently ported from a more FORTRAN-like implementation so this is still required by some of them
            #   for now.
            
            xmap : np.ndarray,
            #   Functional derivatives of the state vector w.r.t Atmospheric profiles at each Atmosphere location.
            #   The array is sized as:
            #    
            #       nx - number of state vector entries.
            #    
            #       NVMR - number of gas volume mixing ratio profiles in the Atmosphere component of the forward model.
            #    
            #       NDUST - number of aerosol profiles in the Atmosphere component of the forward model.
            #    
            #       NP - number of points in an atmospheric profile, all profiles in an Atmosphere component of the forward model 
            #               should have the same number of points.
            #    
            #       NLOCATIONS - number of locations defined in the Atmosphere component of the forward model.
            #
            #   The size of the 1st dimension (NVMR+2+NDUST) is like that because it packs in 4 different atmospheric profile
            #   types: gas volume mixing ratios (NVMR), aerosol densities (NDUST), fractional cloud cover (1), para H2 fraction (1).
        ) -> None:
        """
            Patches values of components based upon values of model parameters in the state vector. Called from ForwardModel_0::subprofretg.
            
            ## ARGUMENTS ##
            
                forward_model : ForwardModel_0
                    The ForwardModel_0 instance that is calling this function. We need this so we can alter components of the forward model
                    inside this function.
                
                ix : int
                    The index of the state vector that corresponds to the start of the model's parameters
                    
                ipar : int
                    An integer that encodes which part of the atmospheric component of the forward model this model should alter. Only
                    used for some Atmospheric models.
                
                ivar : int
                    The model index, the order in which the models were instantiated. NOTE: this is a vestige from the
                    FORTRAN version of the code, we don't really need to know this as we should be: 1) storing any model-specific
                    values on the model instance itself; 2) passing any model-specific data from the outside directly
                    instead of having the model instance look it up from a big array. However, the code for each model
                    was recently ported from a more FORTRAN-like implementation so this is still required by some of them
                    for now.
                
                xmap : np.ndarray[[nx,NVMR+2+NDUST,NP,NLOCATIONS],float]
                    Functional derivatives of the state vector w.r.t Atmospheric profiles at each Atmosphere location.
                    The array is sized as:
                        
                        nx - number of state vector entries.
                        
                        NVMR - number of gas volume mixing ratio profiles in the Atmosphere component of the forward model.
                        
                        NDUST - number of aerosol profiles in the Atmosphere component of the forward model.
                        
                        NP - number of points in an atmospheric profile, all profiles in an Atmosphere component of the forward model 
                             should have the same number of points.
                        
                        NLOCATIONS - number of locations defined in the Atmosphere component of the forward model.
                    
                    The size of the 1st dimension (NVMR+2+NDUST) is like that because it packs in 4 different atmospheric profile
                    types: gas volume mixing ratios (NVMR), aerosol densities (NDUST), fractional cloud cover (1), para H2 fraction (1).
                    
                
            ## RETURNS ##
            
                None
        """
        ...
    
    @abc.abstractmethod
    def calculate_from_subspecret(
            self,
            forward_model : "ForwardModel_0",
            #   The ForwardModel_0 instance that is calling this function. We need this so we can alter components of the forward model
            #   inside this function.
            
            ix : int,
            #   The index of the state vector that corresponds to the start of the model's parameters
            
            ivar : int,
            #   The model index, the order in which the models were instantiated. NOTE: this is a vestige from the
            #   FORTRAN version of the code, we don't really need to know this as we should be: 1) storing any model-specific
            #   values on the model instance itself; 2) passing any model-specific data from the outside directly
            #   instead of having the model instance look it up from a big array. However, the code for each model
            #   was recently ported from a more FORTRAN-like implementation so this is still required by some of them
            #   for now.
            
            SPECMOD : np.ndarray[['NCONV','NGEOM'],float],
            #   Modelled spectrum that we want to alter with this model. NOTE: do not assign directly to this, always
            #   assign to slices of it. If you assign directly, then only the **reference** will change and the value
            #   outside the function will not be altered.
            #
            #   The shape is defined as:
            #
            #       NCONV - number of "convolution points" (i.e. wavelengths/wavenumbers) in the modelled spectrum.
            #
            #       NGEOM - number of "geometries" (i.e. different observation setups) in the modelled spectrum.
            
            dSPECMOD : np.ndarray[['NCONV','NGEOM','NX'],float],
            #   Gradients of the spectrum w.r.t each entry of the state vector.NOTE: do not assign directly to this, always
            #   assign to slices of it. If you assign directly, then only the **reference** will change and the value
            #   outside the function will not be altered.
            #
            #   The shape is defined as:
            #
            #       NCONV - number of "convolution points" (i.e. wavelengths/wavenumbers) in the modelled spectrum.
            #
            #       NGEOM - number of "geometries" (i.e. different observation setups) in the modelled spectrum.
            #
            #       NX - Number of entries in the state vector.
            
            
            
        ) -> None:
        """
            Updated spectra based upon values of model parameters in the state vector. Called from ForwardModel_0::subspecret.
        """
        ...


from __future__ import annotations #  for 3.9 compatability
from enum import IntEnum, IntFlag, auto

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)

class PlanetEnum(IntEnum):
    """
    Define values for 'IPLANET', the planet being observed.
    """
    UNDEFINED = -1
    Mercury = 1
    Venus = 2
    Earth = 3
    Mars = 4
    Jupiter = 5
    Saturn = 6
    Uranus = 7
    Neptune = 8
    Pluto = 9
    Sun = 10
    Titan = 11
    NGTS_10b = 85
    WASP_43b = 87
    

class AtmosphericProfileFormatEnum(IntEnum):
    """
    Defines values for 'AMFORM', the atmospheric profile format.
    """
    MOLECULAR_WEIGHT_DEFINED = 0
    CALC_MOLECULAR_WEIGHT_SCALE_VMR_TO_ONE = 1
    CALC_MOLECULAR_WEIGHT_DO_NOT_SCALE_VMR = 2

class ParaH2Ratio(IntEnum):
    """
    Defines values for 'INORMAL' and elements of 'INORMALT', the para-hydrogen ratio in the atmosphere.
    """
    EQUILIBRIUM = 0 # 1:1 ratio
    NORMAL = 1 # 3:1 ratio

class Gas(IntEnum):
    """
    Enum to define the gases used in the retrievals.
    
    Used in 'IPAIRG1', 'IPAIRG2'
    """
    H2O = 1
    CO2 = 2
    O3 = 3
    N2O = 4
    CO = 5
    CH4 = 6
    O2 = 7
    NO = 8
    SO2 = 9
    NO2 = 10
    NH3 = 11
    HNO3 = 12
    OH = 13
    HF = 14
    HCl = 15
    HBr = 16
    HI = 17
    ClO = 18
    OCS = 19
    H2CO = 20
    HOCl = 21
    N2 = 22
    HCN = 23
    CH3Cl = 24
    H2O2 = 25
    C2H2 = 26
    C2H6 = 27
    PH3 = 28
    C2N2 = 29
    C4H2 = 30
    HC3N = 31
    C2H4 = 32
    GeH4 = 33
    C3H8 = 34
    HCOOH = 35
    H2S = 36
    COF2 = 37
    SF6 = 38
    H2 = 39
    He = 40
    AsH3 = 41
    C3H4 = 42
    ClONO2 = 43
    HO2 = 44
    O = 45
    NO_PLUS = 46
    CH3OH = 47
    H = 48
    C6H6 = 49
    CH3CN = 50
    CH2NH = 51
    C2H3CN = 52
    HCP = 53
    CS = 54
    HC5N = 55
    HC7N = 56
    C2H5CN = 57
    CH3NH2 = 58
    HNC = 59
    Na = 60
    K = 61
    TiO = 62
    VO = 63
    CH2CCH2 = 64
    C4N2 = 65
    C5H5N = 66
    C5H4N2 = 67
    C7H8 = 68
    C8H6 = 69
    C5H5CN = 70
    HOBr = 71
    CH3Br = 72
    CF4 = 73
    SO3 = 74
    Ne = 75
    Ar = 76
    COCl2 = 77
    SO = 78
    H2SO4 = 79
    E_MINUS = 80
    H3_PLUS = 81
    FeH = 82
    AlO = 83
    AlCl = 84
    AlF = 85
    AlH = 86
    BeH = 87
    C2 = 88
    CaF = 89
    CaH = 90
    H_MINUS = 91
    CaO = 92
    CH = 93
    CH3 = 94
    CH3F = 95
    CN = 96
    CP = 97
    CrH = 98
    HD_PLUS = 99
    HeH_PLUS = 100
    KCl = 101
    KF = 102
    LiCl = 103
    LiF = 104
    LiH = 105
    LiH_PLUS = 106
    MgF = 107
    MgH = 108
    MgO = 109
    NaCl = 110
    NaF = 111
    NaH = 112
    NH = 113
    NS = 114
    OH_PLUS = 115
    CIS_P2H2 = 116
    TRANS_P2H2 = 117
    PH = 118
    PN = 119
    PO = 120
    PS = 121
    ScH = 122
    SH = 123
    SiH = 124
    SiH2 = 125
    SiH4 = 126
    SiO = 127
    SiS = 129
    TiH = 130
    Cl2 = 131
    ClO2 = 132
    BrO = 133
    IO = 134
    NO3 = 135
    BrO2 = 136
    IO2 = 137

class InstrumentLineshape(IntEnum):
    """
    Defines values for 'ISHAPE'
    """
    Square = 0
    Triangular = 1
    Gaussian = 2
    Hamming = 3
    Hanning = 4

class WaveUnit(IntEnum):
    """
    Defines values for 'ISPACE'
    """
    Wavenumber_cm = 0
    Wavelength_um = 1

class SpectraUnit(IntEnum):
    """
    Defines values for 'IFORM'
    """
    Radiance = 0 # W cm-2 sr-1 (cm-1)-1 if ISPACE=0 ---- W cm-2 sr-1 μm-1 if ISPACE=1
    FluxRatio = 1 # F_planet/F_star - Dimensionless
    A_Ratio = 2 # A_planet/A_star - 100.0 * A_planet/A_star (dimensionless)
    Integrated_spectral_power = 3 # Integrated spectral power of planet - W (cm-1)-1 if ISPACE=0 ---- W um-1 if ISPACE=1
    Atmospheric_transmission = 4 # Atmospheric transmission multiplied by solar flux
    Normalised_radiance = 5 # Normalised radiance to a given wavelength (VNORM)
    Integrated_radiance = 6 # Integrated radiance over filter function - W cm-2 sr-1 if ISPACE=0 ---- W cm-2 sr-1 if ISPACE=1

class SpectralCalculationMode(IntEnum):
    """
    Defines values for 'ILBL'
    """
    K_TABLES = 0 # use pre-tabulated correlated k-tables
    LINE_BY_LINE_RUNTIME = 1 # calculate line-by-line during runtime
    LINE_BY_LINE_TABLES = 2 # use pre-tabulated line-by-line tables

class LowerBoundaryCondition(IntEnum):
    """
    Defines values for 'LOWBC'
    
    Used in 'Surface_0.py'
    """
    THERMAL = 0 # Thermal emission only (i.e. no reflection)
    LAMBERTIAN = 1 # Lambertian surface
    HAPKE = 2 # Hapke surface
    OREN_NAYAR = 3 # Oren-Nayar surface

class RetrievalStrategy(IntEnum):
    """
    Defines values for 'retrieval_method'
    """
    Optimal_Estimation = 0
    Nested_Sampling = 1

class ScatteringCalculationMode(IntEnum):
    """
    Defines values for 'ISCAT'
    
    Used in 'Scatter_0.py'
    """
    THERMAL_EMISSION = 0 # Thermal emission only (i.e. no scattering)
    MULTIPLE_SCATTERING = 1 # Multiple scattering
    INTERNAL_RADIATION_FIELD = 2 # Internal radiation field
    SINGLE_SCATTERING_PLANE_PARALLEL = 3 # Single scattering in a plane parallel atmosphere
    SINGLE_SCATTERING_SPHERICAL = 4 # Single scattering in a spherical atmosphere
    INTERNAL_NET_FJLUX = 5 # Internal net flux
    DOWNWARD_BOTTOM_FLUX = 6 # Downward flux at the bottom of the atmosphere

class RayleighScatteringMode(IntEnum):
    """
    Defines values for 'IRAY'
    
    Used in 'Scatter_0.py'
    """
    NOT_INCLUDED = 0 # Rayleigh scattering optical depth not included
    GAS_GIANT_ATM = 1 # Rayleigh scattering for gas giant atmospheres
    C02_DOMINATED_ATM = 2 # Rayleigh scattering for CO2 dominated atmospheres
    N2_O2_DOMINATED_ATM = 3 # Rayleigh scattering for N2-O2 dominated atmospheres
    JOVIAN_AIR = 4 # Rayleigh scattering for Jovian air (adaptive from Larry Sromovsky)

class AerosolPhaseFunctionCalculationMode(IntEnum):
    """
    Defines values for 'IMIE'
    
    Used in 'Scatter_0.py'
    """
    HENYEY_GREENSTEIN = 0 # Henyey-Greenstein parameters
    MIE_THEORY = 1 # Explicitly calculated from Mie theory
    LEGENDRE_POLYNOMIALS = 2 # Legendre polynomials

class ZenithAngleOrigin(IntEnum):
    """
    Defines values for 'IPZEN'

    Used in 'AtmCalc_0.py'
    """
    BOTTOM = 0 # Zenith angle is defined at the bottom of the bottom layer
    ALTITUDE_ZERO = 1 # Zenith angle is defined at 0km atltitude
    TOP = 2 # Zenith angle is defined at the top of the top layer


class PathObserverPointing(IntEnum):
    """
    Defines location of the PathObserver, used when calculating radiative transfer.

    Used in 'AtmCalc_0.py'
    """
    LIMB = 0 # Limb path, path observer is looking at the limb of the planet
    NADIR = 1 # Nadir path, path observer is on the planet looking upwards
    DISK = 2 # Disk path, path observer is looking at the disk of the planet


class PathCalc(IntFlag):
    """
    Defines path calculation type used when calculating radiative transfer.
    
    Used as elements of 'IMOD' in 'AtmCalc_0.py', 'Path_0.py', 'ForwardModel_0.py'
    """
    WEIGHTING_FUNCTION = auto() # Weighting function
    NET_FLUX = auto() # Net flux calculation
    UPWARD_FLUX = auto() # Internal upward flux calculation
    OUTWARD_FLUX = auto() # Upward flux at top of topmost layer
    DOWNWARD_FLUX = auto() # Downward flux at bottom of lowest layer
    CURTIS_GODSON = auto() # Curtis Godson
    THERMAL_EMISSION = auto() # Thermal emission
    HEMISPHERE = auto() # Integrate emission into hemisphere
    MULTIPLE_SCATTERING = auto() # Full scattering calculation
    NEAR_LIMB = auto() # Near-limb scattering calculation
    SINGLE_SCATTERING_PLANE_PARALLEL = auto() # Single scattering calculation (plane parallel)
    SINGLE_SCATTERING_SPHERICAL = auto() # Single scattering calculation (spherical atm.)
    ABSORBTION = auto() # calculate absorption not transmission
    PLANCK_FUNCTION_AT_BIN_CENTRE = auto() # use planck function at bin centre in genlbl (also denoted as BINBB in old code)
    BROADENING = auto() # calculate emission outside of genlbl


class LayerType(IntEnum):
    """
    Defines layer type used when calculating radiative transfer.
    
    Used as 'LAYTYP' in 'Layer_0.py'
    """
    EQUAL_PRESSURE = 0
    EQUAL_LOG_PRESSURE = 1
    EQUAL_HEIGHT = 2
    EQUAL_PATH_LENGTH = 3
    BASE_PRESSURE = 4
    BASE_HEIGHT = 5

class LayerIntegrationScheme(IntEnum):
    """
    Defines layer integration scheme used when calculating radiative transfer.
    
    Used as 'LAYINT' in 'Layer_0.py'
    """
    MID_PATH = 0
    ABSORBER_WEIGHTED_AVERAGE = 1

class InterpolationMethod(IntEnum):
    """
    Defines interpolation method used by SCIPY routines.
    
    Used in 'Layer_0.py'
    """
    LINEAR = 0
    QUADRATIC_SPLINE = 1
    CUBIC_SPLINE = 2


class AtmosphericProfileType(IntEnum):
    """
    Defines the atmospheric profile type that a model parameterises
    
    Used as 'ipar' in 'Models.py', 'ForwardModel_0.py'
    """
    NOT_PRESENT = -1
    GAS_VOLUME_MIXING_RATIO = 0
    TEMPERATURE = 1
    AEROSOL_DENSITY = 2
    PARA_H2_FRACTION = 3
    FRACTIONAL_CLOUD_COVERAGE = 4


class SpectroscopicLineList(IntEnum):
    """
    Defines the spectroscopic line list to be used 

    Used as DATABASE in 'LineData_0.py'
    """
    CUSTOM = 0
    HITRAN = 1
    
class AmbientGas(IntEnum):
    """
    Defines the ambient gas used in the line spectroscopy calculations

    Used as ambient_gas in 'LineData_0.py'
    """
    AIR = 0
    CO2 = 1
    H2 = 2


"""
Configures logging for the package
"""
import logging

pkg_lgr = logging.getLogger(__name__.split('.',1)[0])
pkg_lgr.propagate = False

pkg_lgr.setLevel(logging.INFO)

pkg_stream_hdlr = logging.StreamHandler()
pkg_stream_hdlr.setLevel(logging.INFO)

pkg_stream_hdlr_formatter = logging.Formatter('%(levelname)s :: %(funcName)s :: %(filename)s-%(lineno)d :: %(message)s')
pkg_stream_hdlr.setFormatter(pkg_stream_hdlr_formatter)


pkg_lgr.addHandler(pkg_stream_hdlr)

_logger_levels : dict[str,list[int]] = dict()


def get_all_logger_decendents(_lgr):
    """
    Generator that iterates over all decendents of `_lgr` including `_lgr` itself.
    """
    yield _lgr
    
    for name, child_lgr in ((name, l) for name, l in logging.root.manager.loggerDict.items() if ((not isinstance(l, logging.PlaceHolder)) and name.startswith(_lgr.name) and (len(name[len(_lgr.name):].split('.'))>=2))) :
        yield child_lgr

def set_packagewide_level(log_level : int, mode : str = 'exact', _lgr : logging.Logger = pkg_lgr):
    """
    Sets the logging level for the whole package at once.
    
    ## ARGUMENTS ##
        log_level : int
            Level (e.g. logging.DEBUG) that all loggers in the package should be set to
        
        mode : str{'exact', 'min', 'max'} = 'exact'
            How the level should be set. 
                'exact' - set all loggers to the passed `log_level`
                'min' - set all loggers to have a most the passed `log_level`
                'max' - set all loggers to have at least the passed `log_level`
        
        _lgr : logging.Logger = pkg_lgr
            The highest logger in the logging hierarchy that will be affected.
    
    ## RETURNS ##
        None
    """
    for decendent_lgr in get_all_logger_decendents(_lgr):
    
        if mode == 'exact':
            #print(f'Setting logger "{decendent_lgr.name}" to {log_level}')
            decendent_lgr.setLevel(log_level)
        elif mode == 'max':
            if decendent_lgr.level > log_level:
                #print(f'Setting logger "{decendent_lgr.name}" to {log_level}')
                decendent_lgr.setLevel(log_level)
        elif mode == 'min':
            if decendent_lgr.level < log_level:
                #print(f'Setting logger "{decendent_lgr.name}" to {log_level}')
                decendent_lgr.setLevel(log_level)
        else:
            raise ValueError(f'{__name__}.set_packagewide_level(...): Unknown mode "{mode}", should be one of ("exact", "min", "max")')


def push_packagewide_level(log_level : int, mode : str = 'exact', _lgr : logging.Logger = pkg_lgr):
    """
    Sets the logging level for the whole package at once.
    
    ## ARGUMENTS ##
        log_level : int
            Level (e.g. logging.DEBUG) that all loggers in the package should be set to
        
        mode : str{'exact', 'min', 'max'} = 'exact'
            How the level should be set. 
                'exact' - set all loggers to the passed `log_level`
                'min' - set all loggers to have a most the passed `log_level`
                'max' - set all loggers to have at least the passed `log_level`
        
        _lgr : logging.Logger = pkg_lgr
            The highest logger in the logging hierarchy that will be affected.
    
    ## RETURNS ##
        None
    """
    #global _logger_levels
    
    for decendent_lgr in get_all_logger_decendents(_lgr):
        level_stack = _logger_levels.get(decendent_lgr.name, list())
        level_stack.append(decendent_lgr.level)
        _logger_levels[decendent_lgr.name] = level_stack
        
        if mode == 'exact':
            #print(f'Setting logger "{decendent_lgr.name}" to {log_level}')
            decendent_lgr.setLevel(log_level)
        elif mode == 'max':
            if decendent_lgr.level > log_level:
                #print(f'Setting logger "{decendent_lgr.name}" to {log_level}')
                decendent_lgr.setLevel(log_level)
        elif mode == 'min':
            if decendent_lgr.level < log_level:
                #print(f'Setting logger "{decendent_lgr.name}" to {log_level}')
                decendent_lgr.setLevel(log_level)
        else:
            raise ValueError(f'{__name__}.set_packagewide_level(...): Unknown mode "{mode}", should be one of ("exact", "min", "max")')


def pop_packagewide_level(_lgr : logging.Logger = pkg_lgr) -> None:
    #global _logger_levels
    
    for decendent_lgr in get_all_logger_decendents(_lgr):
    
        level_stack = _logger_levels.get(decendent_lgr.name, list())
        if len(level_stack) != 0:
            decendent_lgr.setLevel(level_stack.pop(-1))
            _logger_levels[decendent_lgr.name] = level_stack


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# Stellar_0.py - Object to represent the stellar spectrum.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations #  for 3.9 compatability

#from archnemesis import *
from archnemesis.enums import WaveUnit
import numpy as np
import matplotlib.pyplot as plt
import h5py

from archnemesis.helpers import h5py_helper

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)

#!/usr/local/bin/python3
# -*- coding: utf-8 -*-

###############################################################################################

"""
Created on Tue Jul 22 17:27:12 2021

@author: juanalday

Stellar Class.
"""

class Stellar_0:

    def __init__(self, SOLEXIST=True, DIST=None, RADIUS=None, ISPACE=None, NWAVE=None):

        """
        Attributes
        ----------
        @attribute SOLEXIST: log,
            Flag indicating whether sunlight needs to be included (SOLEXIST=True) or not (SOLEXIST=False)
        @attribute DIST: float,
            Distance between star and planet (AU) 
        @attribute RADIUS: float,
            Radius of the star (km)       
        @attribute ISPACE: int,
            Spectral units in which the solar spectrum is defined (0) Wavenumber (1) Wavelength              
        @attribute NWAVE: int,
            Number of spectral points in which the stellar spectrum is defined
        @attribute WAVE: 1D array
            Wavelength array at which the stellar file is defined
        @attribute SOLSPEC: 1D array
            Stellar luminosity spectrum (W/(cm-1) or W/um)
        @attribute SOLFLUX: 1D array
            Stellar flux at planet's distance (W cm-2 um-1 or W cm-2 (cm-1)-1)
        @attribute STELLARDATA: str
            String indicating where the STELLAR data files are stored

        Methods
        -------
        Stellar_0.assess
        Stellar_0.edit_WAVE
        Stellar_0.edit_SOLSPEC
        Stellar_0.read_hdf5
        Stellar_0.write_hdf5
        Stellar_0.read_sol
        Stellar_0.write_sol
        Stellar_0.calc_solar_flux
        Stellar_0.calc_solar_power
        Stellar_0.write_solar_file
        """

        from archnemesis.Data.path_data import archnemesis_path

        #Input parameters
        self.SOLEXIST = SOLEXIST
        self.DIST = DIST
        self.RADIUS = RADIUS
        #self.ISPACE = WaveUnit(ISPACE) if ISPACE is not None and not isinstance(ISPACE, WaveUnit) else ISPACE
        self.NWAVE = NWAVE

        # Input the following profiles using the edit_ methods.
        self.WAVE = None # np.zeros(NWAVE)
        self.SOLSPEC = None # np.zeros(NWAVE)
        self.SOLFLUX = None #np.zeros(NWAVE)

        self.STELLARDATA = archnemesis_path()+'archnemesis/Data/stellar/'
        
        # private attributes
        self._ispace = None
        
        # set property values
        self.ISPACE = ISPACE if ISPACE is not None else WaveUnit.Wavenumber_cm
    
    
    @property
    def ISPACE(self) -> WaveUnit:
        return self._ispace
    
    @ISPACE.setter
    def ISPACE(self, value):
        self._ispace = WaveUnit(value)


    def assess(self):
        """
        Assess whether the different variables have the correct dimensions and types
        """

        if self.SOLEXIST is True:

            #Checking some common parameters to all cases
            assert isinstance(self.ISPACE, (int, np.integer, WaveUnit)), \
                'ISPACE must be int or WaveUnit'
            assert self.ISPACE in WaveUnit, \
                f'ISPACE must be one of {tuple(WaveUnit)}'

            #Checking some common parameters to all cases
            assert np.issubdtype(type(self.DIST), np.float64) == True , \
                'DIST must be float'

            #Checking some common parameters to all cases
            assert np.issubdtype(type(self.RADIUS), np.float64) == True , \
                'RADIUS must be float'

            #Checking some common parameters to all cases
            assert np.issubdtype(type(self.NWAVE), np.integer) == True , \
                'NWAVE must be int'
            assert self.NWAVE >= 0 , \
                'NWAVE must be >=0'
             
            assert len(self.WAVE) == self.NWAVE , \
                'WAVE must have size (NWAVE)'
            
            assert len(self.SOLSPEC) == self.NWAVE , \
                'SOLSPEC must have size (NWAVE)'
            

    def edit_WAVE(self, WAVE):
        """
        Edit the wavelength array 
        @param WAVE: 1D array
            Array defining the wavelengths at which the solar spectrum is defined
        """
        WAVE_array = np.array(WAVE)
        assert len(WAVE_array) == self.NWAVE, 'WAVE should have NWAVE elements'
        self.WAVE = WAVE_array

    def edit_SOLSPEC(self, SOLSPEC):
        """
        Edit the solar spectrum 
        @param SOLSPEC: 1D array
            Array defining the solar spectrum
        """
        SOLSPEC_array = np.array(SOLSPEC)
        assert len(SOLSPEC_array) == self.NWAVE, 'SOLSPEC should have NWAVE elements'
        self.SOLSPEC = SOLSPEC_array

    def write_hdf5(self,runname,solfile=None):
        """
        Write the information about the solar spectrum in the HDF5 file

        If the optional input solfile is defined, then the information is read from the
        specified file (assumed to be stored in the Data/stellar directory).

        If solfile is not defined, then the information about the solar spectrum is assumed
        to be defined in the class
        """

        from archnemesis.Files import file_lines

        if solfile is not None:

            #Reading the solar spectrum file

            nlines = file_lines(self.STELLARDATA+solfile)

            #Reading buffer
            ibuff = 0
            with open(self.STELLARDATA+solfile,'r') as fsol:
                for curline in fsol:
                    if curline.startswith("#"):
                        ibuff = ibuff + 1
                    else:
                        break

            nvsol = nlines - ibuff - 2
            
            #Reading file
            fsol = open(self.STELLARDATA+solfile,'r')
            for i in range(ibuff):
                s = fsol.readline().split()
        
            s = fsol.readline().split()
            ispace = WaveUnit(int(s[0]))
            s = fsol.readline().split()
            solrad = float(s[0])
            vsol = np.zeros(nvsol)
            rad = np.zeros(nvsol)
            for i in range(nvsol):
                s = fsol.readline().split()
                vsol[i] = float(s[0])
                rad[i] = float(s[1])
        
            fsol.close()


            self.RADIUS = solrad
            self.ISPACE = ispace
            self.NWAVE = nvsol
            self.edit_WAVE(vsol)
            self.edit_SOLSPEC(rad)


        self.assess()

        #Writing the information into the HDF5 file
        with h5py.File(runname+'.h5','a') as f:
            #Checking if Stellar already exists
            if ('/Stellar' in f)==True:
                del f['Stellar']   #Deleting the Stellar information that was previously written in the file

            if self.SOLEXIST is True:

                grp = f.create_group("Stellar")

                #Writing the spectral units
                dset = h5py_helper.store_data(grp, 'ISPACE', int(self.ISPACE))
                dset.attrs['title'] = "Spectral units"
                if self.ISPACE == WaveUnit.Wavenumber_cm:
                    dset.attrs['units'] = 'Wavenumber / cm-1'
                elif self.ISPACE == WaveUnit.Wavelength_um:
                    dset.attrs['units'] = 'Wavelength / um'

                #Writing the Planet-Star distance
                dset = h5py_helper.store_data(grp, 'DIST', self.DIST)
                dset.attrs['title'] = "Planet-Star distance"
                dset.attrs['units'] = 'Astronomical Units'

                #Writing the Star radius
                dset = h5py_helper.store_data(grp, 'RADIUS', self.RADIUS)
                dset.attrs['title'] = "Star radius"
                dset.attrs['units'] = 'km'

                #Writing the number of points in stellar spectrum
                dset = h5py_helper.store_data(grp, 'NWAVE', self.NWAVE)
                dset.attrs['title'] = "Number of spectral points in stellar spectrum"

                #Writing the spectral array
                dset = h5py_helper.store_data(grp, 'WAVE', self.WAVE)
                dset.attrs['title'] = "Spectral array"
                if self.ISPACE == WaveUnit.Wavenumber_cm:
                    dset.attrs['units'] = 'Wavenumber / cm-1'
                elif self.ISPACE == WaveUnit.Wavelength_um:
                    dset.attrs['units'] = 'Wavelength / um' 

                #Writing the solar spectrum
                dset = h5py_helper.store_data(grp, 'SOLSPEC', self.SOLSPEC)
                dset.attrs['title'] = "Stellar power spectrum"
                if self.ISPACE == WaveUnit.Wavenumber_cm:
                    dset.attrs['units'] = 'W (cm-1)-1'
                elif self.ISPACE == WaveUnit.Wavelength_um:
                    dset.attrs['units'] = 'W um-1'     


    def read_hdf5(self,runname):
        """
        Read the Stellar properties from an HDF5 file
        """

        with h5py.File(runname+'.h5','r') as f:

            #Checking if Surface exists
            e = "/Stellar" in f
            if e is False:
                self.SOLEXIST = False
            else:
                self.SOLEXIST = True
                self.ISPACE = h5py_helper.retrieve_data(f, 'Stellar/ISPACE', lambda x:  WaveUnit(np.int32(x)))
                self.DIST = h5py_helper.retrieve_data(f, 'Stellar/DIST', np.float64)
                self.RADIUS = h5py_helper.retrieve_data(f, 'Stellar/RADIUS', np.float64)
                self.NWAVE = h5py_helper.retrieve_data(f, 'Stellar/NWAVE', np.int32)
                self.WAVE = h5py_helper.retrieve_data(f, 'Stellar/WAVE', np.array)
                self.SOLSPEC = h5py_helper.retrieve_data(f, 'Stellar/SOLSPEC', np.array)


    def read_sol(self, runname, MakePlot=False):
        """
        Read the solar spectrum from the .sol file. There are two options for this file:

            - The only line in the file is the name of another file including the solar power spectrum,
              assumed to be stored in the Data/stellar/ directory

            - The first line is equal to -1. Then the stellar spectrum is read from the solar file.

        @param runname: str
            Name of the NEMESIS run
        """

        from archnemesis.Files import file_lines

        #Opening file
        f = open(runname+'.sol','r')
        s = f.readline().split()
        solname = s[0]
        f.close()

        if solname=='-1':
            #Information about stellar spectrum is stored in this same file
            nlines = file_lines(runname+'.sol')
            nvsol = nlines - 3


            f = open(runname+'.sol','r')
            s = f.readline().split()
            solname = s[0]

            s = f.readline().split()
            ispace = WaveUnit(int(s[0]))
            s = f.readline().split()
            solrad = float(s[0])
            vsol = np.zeros(nvsol)
            rad = np.zeros(nvsol)
            for i in range(nvsol):
                s = f.readline().split()
                vsol[i] = float(s[0])
                rad[i] = float(s[1])
            f.close()

        else:

            nlines = file_lines(self.STELLARDATA+solname)

            #Reading buffer
            ibuff = 0
            with open(self.STELLARDATA+solname,'r') as fsol:
                for curline in fsol:
                    if curline.startswith("#"):
                        ibuff = ibuff + 1
                    else:
                        break

            nvsol = nlines - ibuff - 2
            
            #Reading file
            fsol = open(self.STELLARDATA+solname,'r')
            for i in range(ibuff):
                s = fsol.readline().split()
        
            s = fsol.readline().split()
            ispace = WaveUnit(int(s[0]))
            s = fsol.readline().split()
            solrad = float(s[0])
            vsol = np.zeros(nvsol)
            rad = np.zeros(nvsol)
            for i in range(nvsol):
                s = fsol.readline().split()
                vsol[i] = float(s[0])
                rad[i] = float(s[1])
        
            fsol.close()

        self.RADIUS = solrad
        self.ISPACE = ispace
        self.NWAVE = nvsol
        self.edit_WAVE(vsol)
        self.edit_SOLSPEC(rad)

        if MakePlot is True:
            fig,ax1=plt.subplots(1,1,figsize=(8,3))
            ax1.plot(vsol,rad)
            #ax1.set_yscale('log')
            plt.tight_layout()
            plt.show()

    def write_sol(self,runname):
        """
        Write the solar power into a .sol file with the format required by NEMESISpy, in the case
        that the solar power spectrum is stored directy in the .sol file
        """

        f = open(runname+'.sol','w')

        #Defining errors while writing file
        if self.ISPACE is None:
            raise ValueError('error :: ISPACE must be defined in Stellar class to write Stellar power to file')

        if self.RADIUS is None:
            raise ValueError('error :: RADIUS must be defined in Stellar class to write Stellar power to file')

        if self.NWAVE is None:
            raise ValueError('error :: NWAVE must be defined in Stellar class to write Stellar power to file')

        if self.WAVE is None:
            raise ValueError('error :: WAVE must be defined in Stellar class to write Stellar power to file')

        if self.SOLSPEC is None:
            raise ValueError('error :: SOLSPEC Must be defined in Stellar class to write Stellar power to file')


        header = '-1'
        f.write(header+' \n')
        f.write('\t %i \n' % (int(self.ISPACE)))
        f.write('\t %7.3e \n' % (self.RADIUS))
        for i in range(self.NWAVE):
            f.write('\t %7.6f \t %7.5e \n' % (self.WAVE[i],self.SOLSPEC[i]))
        f.close()


    def calc_solar_flux(self):
        """
        Calculate the stellar flux at the planet's distance (W cm-2 (cm-1)-1 or W cm-2 um-1)
        """

        AU = 1.49598e11
        area = 4.*np.pi*(self.DIST * AU * 100. )**2.
        self.SOLFLUX = self.SOLSPEC / area   #W cm-2 (cm-1)-1 or W cm-2 um-1


    def calc_solar_power(self):
        """
        Calculate the stellar power based on the solar flux measured at a given distance
        """

        AU = 1.49598e11
        area = 4.*np.pi*(self.DIST * AU * 100. )**2.
        self.SOLSPEC = self.SOLFLUX * area   #W (cm-1)-1 or W um-1


    def write_solar_file(self,filename,header=None):
        """
        Write the solar power into a file with the format required by NEMESIS
        """

        f = open(filename,'w')

        #Defining errors while writing file
        if self.ISPACE is None:
            raise ValueError('error :: ISPACE must be defined in Stellar class to write Stellar power to file')

        if self.RADIUS is None:
            raise ValueError('error :: RADIUS must be defined in Stellar class to write Stellar power to file')

        if self.NWAVE is None:
            raise ValueError('error :: NWAVE must be defined in Stellar class to write Stellar power to file')

        if self.WAVE is None:
            raise ValueError('error :: WAVE must be defined in Stellar class to write Stellar power to file')

        if self.SOLSPEC is None:
            raise ValueError('error :: SOLSPEC Must be defined in Stellar class to write Stellar power to file')


        if header is None:
            if self.ISPACE == WaveUnit.Wavenumber_cm:
                header = '# Stellar power in W (cm-1)-1'
            elif self.ISPACE == WaveUnit.Wavelength_um:
                header = '# Stellar power in W um-1' 
        else:
            if header[0]!='#':
                header = '#'+header
        
        f.write(header+' \n')
        f.write('\t %i \n' % (int(self.ISPACE)))
        f.write('\t %7.3e \n' % (self.RADIUS))
        for i in range(self.NWAVE):
            f.write('\t %7.6f \t %7.5e \n' % (self.WAVE[i],self.SOLSPEC[i]))

        f.close()
        
###############################################################################################


import pytest
import archnemesis as ans
import numpy as np

def create_example_surface(NEM, NLOCATIONS):
    surface = ans.Surface_0(
        GASGIANT=True, 
        ISPACE=1, # Spectral units (0 = wavenumber, 1 = um)
        LOWBC=1, # Lower boundary condition
        GALB=-1.0, # Ground albedo
        NEM=NEM, # Number of points in the EMissivity spectrum of the surface
        NLOCATIONS=NLOCATIONS # Number of points "on the surface" for which we define a spectrum/temperature etc.
    )
    
    # Make a grid of LAT-LON values
    nx = int(np.floor(np.sqrt(surface.NLOCATIONS)))
    ny = surface.NLOCATIONS // nx
    nz = surface.NLOCATIONS - nx*ny
    surface.LATITUDE = np.repeat(np.linspace(-90,90,nx), ny)
    surface.LATITUDE = np.concatenate((surface.LATITUDE, np.zeros((nz,))))

    surface.LONGITUDE = np.repeat(np.linspace(-90,90,ny), nx).reshape((ny,nx)).T.flatten()
    surface.LONGITUDE = np.concatenate((surface.LONGITUDE, np.zeros((nz,))))

    surface.TSURF=np.ones((surface.NLOCATIONS,))*300
    
    # Wavelengths where surface parameters are defined (ISPACE = 1, so is in um)
    surface.VEM = np.geomspace(1E-9,100,surface.NEM) 
    
    return surface


@pytest.mark.parametrize(
    'NEM, NLOCATIONS', [
        (201,100), 
        (201,1), 
        (100,201), 
        (1,1),
        (1,2)
    ]
)
def test_surface_radiance_calculation(NEM, NLOCATIONS):
    surface = create_example_surface(NEM, NLOCATIONS)
    
    # inside Surface_0.calc_radground(...), bbsurf has shape (self.NEM, self.NLOCATIONS)
    # so surface.EMISSIVITY must have that shape as well.
    surface.EMISSIVITY = np.ones((surface.NEM, surface.NLOCATIONS))
    
    # To test the 1D case, squeeze the 2nd dimension if we can
    if (surface.EMISSIVITY.shape[1] == 1):
        surface.EMISSIVITY = np.squeeze(surface.EMISSIVITY, axis=1)
    
    # This should not throw an error
    surface_radiance = surface.calc_radground(surface.ISPACE)


#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# planet_data.py - Data reference files for planetary parameters.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Data reference files.

Contains the parameters describing different planets.
"""

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)


planet_info = {
    "1": {
        "name": "Mercury",
        "mass": 0.33022e24,  #kg
        "radius": 2439.7,    #km
        "flatten": 0.0,      
        "rotation": 58.6462, #days
        "isurf": 1,
        "Jcoeff": [0.0,0.0,0.0]
        },
    "2": {
        "name": "Venus",
        "mass": 4.8690,  #kg
        "radius": 6051.8,    #km
        "flatten": 0.0,      
        "rotation": -243.0187, #days
        "isurf": 1,
        "Jcoeff": [0.027,0.0,0.0]
        },
    "3": {
        "name": "Earth",
        "mass": 5.9742,  #kg
        "radius": 6378.14,    #km
        "flatten": 0.00335364,      
        "rotation": 0.99726968, #days
        "isurf": 1,
        "Jcoeff": [1.08263,-2.54,-1.61]
        },
    "4": {
        "name": "Mars",
        "mass": 0.64191,  #kg
        "radius": 3397.0,    #km
        "flatten": 0.006476,      
        "rotation": 1.02595675, #days
        "isurf": 1,
        "Jcoeff": [1.964,36.0,0.0]
        }, 
    "5": {
        "name": "Jupiter",
        "mass": 1898.2,  #kg
        "radius": 71492.0,    #km
        "flatten": 0.064874,      
        "rotation": 0.41354, #days
        "isurf": 0,
        "Jcoeff": [14.75,0.0,-580.0]
        },
    "6": {
        "name": "Saturn",
        "mass": 568.5,  #kg
        "radius": 60268.0,    #km
        "flatten": 0.097962,      
        "rotation": 0.44401, #days
        "isurf": 0,
        "Jcoeff": [16.45,0.0,-1000.0]
        },
    "7": {
        "name": "Uranus",
        "mass": 86.625,  #kg
        "radius": 25559.0,    #km
        "flatten": 0.022000,      
        "rotation": -0.71833, #days
        "isurf": 0,
        "Jcoeff": [12.0,0.0,0.0]
        },
    "8": {
        "name": "Neptune",
        "mass": 102.78,  #kg
        "radius": 24764.0,    #km
        "flatten": 0.017081,      
        "rotation": 0.67125, #days
        "isurf": 0,
        "Jcoeff": [4.0,0.0,0.0]
        },
    "9": {
        "name": "Pluto",
        "mass": 0.015,  #kg
        "radius": 1151.0,    #km
        "flatten": 0.0,      
        "rotation": -6.3872, #days
        "isurf": 1,
        "Jcoeff": [0.0,0.0,0.0]
        },
    "10": {
        "name": "Sun",
        "mass": 1989000.0,  #kg
        "radius": 695000.0,    #km
        "flatten": 0.0,      
        "rotation": 25.38, #days
        "isurf": 0,
        "Jcoeff": [0.0,0.0,0.0]
        },
    "11": {
        "name": "Titan",
        "mass": 0.1353,  #kg
        "radius": 2575.0,    #km
        "flatten": 0.0,      
        "rotation": 15.945, #days
        "isurf": 1,
        "Jcoeff": [0.0,0.0,0.0]
        },
    "85": {
        "name": "NGTS-10b",
        "mass": 4103.757,  #kg
        "radius": 86147.86,    #km
        "flatten": 0.0,      
        "rotation": 100000.0, #days
        "isurf": 0,
        "Jcoeff": [0.0,0.0,0.0]
        },
    "87": {
        "name": "WASP-43b",
        "mass": 3895.110,  #kg
        "radius": 74065.70,    #km
        "flatten": 0.0,      
        "rotation": 100000.0, #days
        "isurf": 0,
        "Jcoeff": [0.0,0.0,0.0]
        }
}


from __future__ import annotations #  for 3.9 compatability

import h5py
from typing import Callable, Any

import numpy as np

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.WARN)

def retrieve_data(
        h5py_file : h5py.File | h5py.Group,
        item_path : str,
        mutator : Callable[[Any], Any] = lambda x: x, # default is identity function
        default : Any = None,
    ) -> Any:
    """
    Retrieves `item_path` data from `h5py_file`, passing it through the `mutator` callable as it does so.
    Makes it easier to ensure we return a certain type from this function but also enables the
    setting of a `default` value for cases where `item_path` is not present in `h5py_file`.
    """
    if item_path in h5py_file and h5py_file[item_path].shape is not None:
        return mutator(h5py_file[item_path])
    else:
        _lgr.warning(f'When reading file "{h5py_file.filename}", could not find element "{item_path}" setting returned value to "{default}"', stacklevel=2)
        return default


def store_data(
        h5py_file : h5py.File | h5py.Group,
        item_path : str,
        data : Any,
        dtype = None, # will guess data type
    ) -> None:
    r"""
    Stores `data` at `item_path` in `h5py_file`. Values of "None" create an empty dataset
    
    Regex replacement for previous version "(\w*?)\.create_dataset\(('.*?'),\s*data\s*=\s*(.*)\)" -> "h5py_helper.store_data($1, $2, $3)"
    """
    #f.create_dataset('Retrieval/Output/OptimalEstimation/NX',data=self.NX)
    
    if dtype is None:
        dtype = float
        if issubclass(type(data), np.ndarray):
            dtype = data.dtype
        elif type(data) is int:
            dtype = int
    
    if item_path not in h5py_file:
        
        if data is not None:
            return h5py_file.create_dataset(item_path, data=data, dtype=dtype)
        else:
            return h5py_file.create_dataset(item_path, shape=None, dtype=dtype)
    
    if data is not None:
        dset = h5py_file[item_path]
        dset[...] = data
        return dset
    else:
        del h5py_file[item_path]
        return h5py_file.create_dataset(item_path, shape=None, dtype=dtype)
        
    
    




#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
#
# archNEMESIS - Python implementation of the NEMESIS radiative transfer and retrieval code
# Retrievals.py - Subroutines to perform atmospheric retrievals.
#
# Copyright (C) 2025 Juan Alday, Joseph Penn, Patrick Irwin,
# Jack Dobinson, Jon Mason, Jingxuan Yang
#
# This file is part of archNEMESIS.
#
# archNEMESIS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations #  for 3.9 compatability
import archnemesis as ans
from archnemesis.enums import RetrievalStrategy
import time
import numpy as np

import logging
_lgr = logging.getLogger(__name__)


def retrieval_nemesis(
        runname,
        legacy_files=False,
        NCores=1,
        retrieval_method : RetrievalStrategy = RetrievalStrategy.Optimal_Estimation,
        nemesisSO=False,
        nemesisdisc=False,
        NS_prefix='chains/'
    ):
    
    """
        FUNCTION NAME : retrieval_nemesis()
        
        DESCRIPTION :
        
            Function to run a NEMESIS retrieval based on the information in the input files
        
        INPUTS :
        
            runname :: Name of the retrieval run (i.e., name of the input files)
        
        OPTIONAL INPUTS:

            legacy_files :: If True, it reads the inputs from the standard Fortran NEMESIS files
                            If False, it reads the inputs from the archNEMESIS HDF5 file
            NCores :: Number of parallel processes for the numerical calculation of the Jacobian
            retrieval_method :: (0) Optimal Estimation formalism
                                (1) Nested sampling
            nemesisSO :: If True, it indicates that the retrieval is a solar occultation observation
            nemesisdisc :: If True, it indicates that the retrieval is a disc-averaged observation
        
        OUTPUTS :
        
            Output files
        
        CALLING SEQUENCE:
        
            retrieval_nemesis(runname,legacy_files=False,NCores=1)
        
        MODIFICATION HISTORY : Juan Alday (21/09/2024)
        
    """ 
    
    start = time.time()

    ######################################################
    ######################################################
    #    READING INPUT FILES AND SETTING UP VARIABLES
    ######################################################
    ######################################################

    if legacy_files is False:
        Atmosphere,Measurement,Spectroscopy,Scatter,Stellar,Surface,CIA,Layer,Variables,Retrieval,Telluric = ans.Files.read_input_files_hdf5(runname)
    else:
        Atmosphere,Measurement,Spectroscopy,Scatter,Stellar,Surface,CIA,Layer,Variables,Retrieval = ans.Files.read_input_files(runname)
        Telluric = None

    ######################################################
    ######################################################
    #    USE INFORMATION FROM PREVIOUS RETRIEVALS 
    ######################################################
    ######################################################

    if Retrieval.LIN>0:

        #Reading .pre file
        Variables_prev = ans.Files.read_pre(runname)
        
        if Retrieval.LIN==1:
            
            _lgr.info('lin=1 :: Using information from previous retrieval to update parameters in reference classes and calculate the error from the previous covariance matrix in new measurement')
            
            #Calculating forward model
            FM_prev = ans.ForwardModel_0(Atmosphere=Atmosphere,Measurement=Measurement,Spectroscopy=Spectroscopy,Scatter=Scatter,Stellar=Stellar,Surface=Surface,CIA=CIA,Layer=Layer,Variables=Variables_prev,Telluric=Telluric)
            YN,KK = FM_prev.jacobian_nemesis(NCores=NCores,nemesisSO=nemesisSO,nemesisdisc=nemesisdisc)
            
            #Calculating forward modelling error
            SF = KK @ Variables_prev.SA @ KK.T

            #Adding forward model error to the new measurement error
            Measurement.SE += SF

            #Updating reference classes with the ones from the forward model class, which have been updated using the previous retrievals
            #Note we do not update the measurement class as that was different in the previous retrieval
            Atmosphere = FM_prev.AtmosphereX
            Spectroscopy = FM_prev.SpectroscopyX
            Scatter = FM_prev.ScatterX
            Stellar = FM_prev.StellarX
            Surface = FM_prev.SurfaceX
            CIA = FM_prev.CIAX
            Layer = FM_prev.LayerX

        elif Retrieval.LIN==2:
            
            _lgr.info('lin=2 :: Using information from previous retrieval to update a priori covariance matrix')
            
            #Reading .pre file
            Variables_prev = ans.Files.read_pre(runname)
            
            ix1 = 0
            for ivar1 in range(Variables.NVAR):
                
                ix2 = 0
                for ivar2 in range(Variables_prev.NVAR):
                    
                    if np.all(Variables.VARIDENT[ivar1, :] == Variables_prev.VARIDENT[ivar2, :]):
                        _lgr.info('Updating variable '+str(Variables.VARIDENT[ivar1,:]))
                        
                        #update things
                        sa_prev = Variables_prev.SA[ix2:ix2+Variables_prev.NXVAR[ivar2],ix2:ix2+Variables_prev.NXVAR[ivar2]]
                        xn_prev = Variables_prev.XN[ix2:ix2+Variables_prev.NXVAR[ivar2]]
                        
                        Variables.SA[ix1:ix1+Variables.NXVAR[ivar1],ix1:ix1+Variables.NXVAR[ivar1]] = sa_prev
                        Variables.XN[ix1:ix1+Variables.NXVAR[ivar1]] = xn_prev
                        Variables.XA[ix1:ix1+Variables.NXVAR[ivar1]] = xn_prev

                    ix2 += Variables_prev.NXVAR[ivar2]
                ix1 += Variables.NXVAR[ivar1]
                    
        elif Retrieval.LIN==3:   

            _lgr.info('lin=3 :: Using information from previous retrieval to update a priori covariance matrix and calculate the new forward model error')
            
            #Reading .pre file
            Variables_prev = ans.Files.read_pre(runname)
            
            #First of all, we update the state vector of the new retrieval with the one from the previous retrieval
            ix1 = 0
            for ivar1 in range(Variables.NVAR):
                
                ix2 = 0
                for ivar2 in range(Variables_prev.NVAR):
                    
                    if np.all(Variables.VARIDENT[ivar1, :] == Variables_prev.VARIDENT[ivar2, :]):
                        _lgr.info('Updating variable '+str(Variables.VARIDENT[ivar1,:]))
                        
                        #update things
                        sa_prev = Variables_prev.SA[ix2:ix2+Variables_prev.NXVAR[ivar2],ix2:ix2+Variables_prev.NXVAR[ivar2]]
                        xn_prev = Variables_prev.XN[ix2:ix2+Variables_prev.NXVAR[ivar2]]
                        
                        Variables.SA[ix1:ix1+Variables.NXVAR[ivar1],ix1:ix1+Variables.NXVAR[ivar1]] = sa_prev
                        Variables.XN[ix1:ix1+Variables.NXVAR[ivar1]] = xn_prev
                        Variables.XA[ix1:ix1+Variables.NXVAR[ivar1]] = xn_prev

                    ix2 += Variables_prev.NXVAR[ivar2]
                ix1 += Variables.NXVAR[ivar1]

            #Now we calculate the forward model error from the previous retrieval
            FM_prev = ans.ForwardModel_0(Atmosphere=Atmosphere,Measurement=Measurement,Spectroscopy=Spectroscopy,Scatter=Scatter,Stellar=Stellar,Surface=Surface,CIA=CIA,Layer=Layer,Variables=Variables_prev,Telluric=Telluric)
            YN,KK = FM_prev.jacobian_nemesis(NCores=NCores,nemesisSO=nemesisSO,nemesisdisc=nemesisdisc)
            
            #We do not want to include the forward model error from variables that are retrieved again now
            ix1 = 0
            for ivar1 in range(Variables.NVAR):
                
                ix2 = 0
                for ivar2 in range(Variables_prev.NVAR):
                    
                    if np.all(Variables.VARIDENT[ivar1, :] == Variables_prev.VARIDENT[ivar2, :]):
                        
                        _lgr.info('Removing forward model error contribution from variable '+str(Variables.VARIDENT[ivar1,:]))
                        
                        #Removing contribution from KK
                        KK[:,ix2:ix2+Variables_prev.NXVAR[ivar2]] = 0.0
            
                    ix2 += Variables_prev.NXVAR[ivar2]
                ix1 += Variables.NXVAR[ivar1]
            
            #Calculating forward modelling error
            SF = KK @ Variables_prev.SA @ KK.T

            #Adding forward model error to the new measurement error
            Measurement.SE += SF

            #Updating reference classes with the ones from the forward model class, which have been updated using the previous retrievals
            #Note we do not update the measurement class as that was different in the previous retrieval
            Atmosphere = FM_prev.AtmosphereX
            Spectroscopy = FM_prev.SpectroscopyX
            Scatter = FM_prev.ScatterX
            Stellar = FM_prev.StellarX
            Surface = FM_prev.SurfaceX
            CIA = FM_prev.CIAX
            Layer = FM_prev.LayerX

    ######################################################
    ######################################################
    #      RUN THE RETRIEVAL USING ANY APPROACH
    ######################################################
    ######################################################

    if retrieval_method == RetrievalStrategy.Optimal_Estimation:
        OptimalEstimation = ans.coreretOE(runname,Variables,Measurement,Atmosphere,Spectroscopy,Scatter,Stellar,Surface,CIA,Layer,Telluric,\
                                          NITER=Retrieval.NITER,PHILIMIT=Retrieval.PHILIMIT,LIN=Retrieval.LIN,NCores=NCores,
                                          nemesisSO=nemesisSO,nemesisdisc=nemesisdisc)
        Retrieval = OptimalEstimation
    elif retrieval_method == RetrievalStrategy.Nested_Sampling:
        from archnemesis.NestedSampling_0 import coreretNS
        
        NestedSampling = coreretNS(runname,Variables,Measurement,Atmosphere,Spectroscopy,Scatter,Stellar,Surface,CIA,Layer,Telluric,NS_prefix=NS_prefix)
        Retrieval = NestedSampling
    else:
        raise ValueError('error in retrieval_nemesis :: Retrieval scheme has not been implemented yet')


    ######################################################
    ######################################################
    #                WRITE OUTPUT FILES
    ######################################################
    ######################################################

    if retrieval_method == RetrievalStrategy.Optimal_Estimation:
        
        if legacy_files is False:
            Retrieval.write_output_hdf5(runname,Variables)
        else:
            Retrieval.write_cov(runname,Variables,pickle=False)
            Retrieval.write_mre(runname,Variables,Measurement)
            Retrieval.write_raw(runname,Variables,Atmosphere)
            
    if retrieval_method == RetrievalStrategy.Nested_Sampling:
        Retrieval.make_plots()

    #Finishing pogram
    end = time.time()
    _lgr.info('Model run OK')
    _lgr.info(' Elapsed time (s) = '+str(end-start))



from __future__ import annotations #  for 3.9 compatability

import io
from typing import NamedTuple, TYPE_CHECKING
import struct


import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

import archnemesis as ans
from archnemesis.database.datatypes.wave_point import WavePoint
from archnemesis.database.datatypes.gas_descriptor import RadtranGasDescriptor

import logging
_lgr = logging.getLogger(__name__)
_lgr.setLevel(logging.DEBUG)

if TYPE_CHECKING:
    NPRESS = "Number of points in pressure grid"
    NTEMP_PER_PRESSURE = "Number of temperatures per pressure point, i.e. number of temperature profiles"
    NTEMP = "Number of points in temperature grid"
    NWAVE = "Number of wave points"


_header_struct = struct.Struct('<2i2f4i')

def _get_body_structs(nwave, npress, ntemp):
    _lgr.debug(f'{nwave=} {npress=} {ntemp=} {nwave * npress * abs(ntemp)=}')
    fortran_record_shape = (
        npress, 
        ntemp if ntemp > 0 else -ntemp*npress, 
        nwave * npress * abs(ntemp)
    )
    return (
        fortran_record_shape,
        struct.Struct(f'<{fortran_record_shape[0]}f'),
        struct.Struct(f'<{fortran_record_shape[1]}f'),
        struct.Struct(f'<{fortran_record_shape[2]}f'),
    )

class LblHeader(NamedTuple):
    irec0 : int # This is ignored and always 9
    nwave : int 
    vmin : float
    delv : float
    npress : int
    ntemp : int
    gas_id : int 
    iso_id : int


class LblDataTProfilesAtPressure(NamedTuple):
    """
    Line-by-line data specified at `NPRESS` pressures with `NTEMP_PER_PRESSURE` temperature profiles for each pressure.
    
    Temperature profiles are associated by index, so the 0th temperature across all pressures is the 0th temperature profile.
    """
    gas_id : int # RADTRAN
    iso_id : int # RADTRAN
    wave_unit : ans.enums.WaveUnit # ENUM value
    wave : np.ndarray[['NWAVE'], float] # in `wave_unit`
    press : np.ndarray[['NPRESS'], float] # bar
    temp : np.ndarray[['NPRESS', 'NTEMP_PER_PRESSURE'], float] # kelvin
    k : np.ndarray[['NWAVE','NPRESS','NTEMP_PER_PRESSURE'], float] # 1E20 (cm^2), see function `Spectroscopy_0::write_lbltable` for precident
    
    def write_legacy_header(self, f : str | io.IOBase):
        if not isinstance(f, io.IOBase):
            with open(f, 'wb') as g:
                return self.write_legacy_header(g)
        
        f.write(_header_struct.pack(
            9 + self.press.size + self.temp.size, # "fortran record number" at which absorption coefficient data starts
            self.wave.size,
            self.wave[0],
            self.wave[1] - self.wave[0],
            self.press.size,
            -self.temp.shape[1],
            self.gas_id,
            self.iso_id,
        ))
    
    def write_legacy(self, f : str | io.IOBase):
        if not isinstance(f, io.IOBase):
            with open(f, 'wb') as g:
                return self.write_legacy(g)
        print(f'Writing {self.__class__.__name__} to {f.name}')
        
        self.write_legacy_header(f)
        
        fortran_record_shape, p_struct, t_struct, k_struct = _get_body_structs(self.wave.size, self.press.size, -self.temp.shape[1])
        
        print(f'{p_struct=}')
        f.write(
            p_struct.pack(
                #*(x for x in np.nditer(np.asfortranarray(self.press)))
                *self.press.flat
            )
        )
        
        print(f'{t_struct=}')
        f.write(
            t_struct.pack(
                #*(x for x in np.nditer(np.asfortranarray(self.temp)))
                *self.temp.flat
            )
        )
        
        print(f'{k_struct=}')
        f.write(
            k_struct.pack(
                #*(x for x in np.nditer(np.asfortranarray(self.k))) # already multiplied by 1E20 to avoid underflow when saving single precision floats
                *self.k.flat
            )
        )
        
        _lgr.info(f'Written {self.__class__.__name__} to {f.name}')
        return
    
    def plot(self, wave_unit : None | ans.enums.WaveUnit = None, z_logscale=True):
        """
        Show a map of `k` (color axis) vs pressure (y-axis) and wave (x-axis) for each temperature 
        profile. Also plots the temperature profile as a red-line on a shared y-axis.
        """
        if wave_unit is None:
            wave_unit = self.wave_unit
        
        fig, ax = plt.subplots(self.k.shape[2],1, figsize=(12,6*self.k.shape[2]), squeeze=False)
        ax = ax.flatten()
        ax2 = [a.twiny() for a in ax]
        for a in ax2:
            a.sharex(ax2[0])
        
        
        def get_edges(a):
            e = np.empty((a.size+1,), dtype=a.dtype)
            half_da = (a[1:] - a[:-1])/2
            e[0] = a[0] - half_da[0]
            e[1:-1] = a[:-1]+half_da
            e[-1] = a[-1] + half_da[-1]
            return e
        
        wave_points = WavePoint(self.wave, self.wave_unit).to_unit(wave_unit)
        print(f'{wave_points=}')
        
        p_edges = np.exp(get_edges(np.log(self.press)))
        w_edges = get_edges(wave_points.value)
        
        fig.suptitle(f'Opacities for {RadtranGasDescriptor(self.gas_id, self.iso_id).label} at each temperature profile.')
        
        for i in range(self.k.shape[2]):
            im = ax[i].pcolormesh(
                w_edges,
                p_edges,
                (np.log(self.k[:,:,i].T)) if z_logscale else (self.k[:,:,i].T)
            )
            
            div = make_axes_locatable(ax[i])
            cax = div.append_axes('right', size=0.05, pad=0.0)
            fig.colorbar(im, cax=cax, orientation='vertical')
            
            if wave_points.unit == ans.enums.WaveUnit.Wavenumber_cm:
                ax[i].set_xlabel('Wavenumber ($cm^{-1}$)')
                if z_logscale:
                    cax.set_ylabel('log[Absorption Coefficient] log(cm$^{-2}$ / $cm^{-1}$)')
                else:
                    cax.set_ylabel('Absorption Coefficient] (cm$^{-2}$ / $cm^{-1}$)')
            elif wave_points.unit == ans.enums.WaveUnit.Wavelength_um:
                ax[i].set_xlabel(r'Wavelength ($\mu$m)')
                if z_logscale:
                    cax.set_ylabel(r'log[Absorption Coefficient] log(cm$^{-2}$ / $\mu$m)')
                else:
                    cax.set_ylabel(r'Absorption Coefficient (cm$^{-2}$ / $\mu$m)')
            else:
                raise RuntimeError(f'Unknown {wave_unit}. Should be one of {ans.enums.WaveUnit.values()}')
            
            
            
            
            
            ax[i].set_ylabel('Pressure (atm)')
            ax[i].set_yscale('log')
            ax[i].invert_yaxis()
            
            
            ax2[i].plot(self.temp[:,i], self.press, color='red', linestyle='--', alpha=0.6)
            ax2[i].set_xlabel('Temperature (K)')
            
            
            
            
            
        
        plt.show()



class LblDataTPGrid(NamedTuple):
    """
    Line-by-line data specified on a 2-dimensional (`NPRESS`,`NTEMP`) grid. All temperature and pressure
    combinations on the grid have a `k` entry.
    """
    gas_id : int # RADTRAN
    iso_id : int # RADTRAN
    wave_unit : ans.enums.WaveUnit # ENUM value
    wave : np.ndarray[['NWAVE'], float] # wavenumber cm^-1
    press : np.ndarray[['NPRESS'], float] # bar
    temp : np.ndarray[['NTEMP'], float] # Kelvin
    k : np.ndarray[['NWAVE','NPRESS','NTEMP'], float] # 1E20 (cm^2), see function `Spectroscopy_0::write_lbltable` for precident


    def write_legacy_header(self, f : str | io.IOBase):
        if not isinstance(f, io.IOBase):
            with open(f, 'wb') as g:
                return self.write_legacy_header(g)
        
        f.write(_header_struct.pack(
            9 + self.press.size + self.temp.size, # "fortran record number" at which absorption coefficient data starts
            self.wave.size,
            self.wave[0],
            self.wave[1] - self.wave[0],
            self.press.size,
            self.temp.size,
            self.gas_id,
            self.iso_id,
        ))
    
    def write_legacy(self, f : str | io.IOBase):
        if not isinstance(f, io.IOBase):
            with open(f, 'wb') as g:
                return self.write_legacy(g)
        
        self.write_legacy_header(f)
        
        fortran_record_shape, p_struct, t_struct, k_struct = _get_body_structs(self.wave.size, self.press.size, self.temp.size)
        
        f.write(
            p_struct.pack(
                *(x for x in np.nditer(np.asfortranarray(self.press)))
            )
        )
        
        f.write(
            t_struct.pack(
                *(x for x in np.nditer(np.asfortranarray(self.temp)))
            )
        )
        
        f.write(
            k_struct.pack(
                *(x for x in np.nditer(np.asfortranarray(self.k))) # already multiplied by 1E20 to avoid underflow when saving single precision floats
            )
        )

    def plot(self, wave_unit : ans.enums.WaveUnit = ans.enums.WaveUnit.Wavelength_um):
        """
        Show a map of `k` (color axis) vs pressure (y-axis) and wave (x-axis) for each temperature
        point on the grid.
        """
        fig, ax = plt.subplots(self.k.shape[2],1, figsize=(12,6*self.k.shape[2]), squeeze=False)
        ax = ax.flatten()
        
        wave_points = WavePoint(self.wave, self.wave_unit).to_unit(wave_unit)
        
        def get_edges(a):
            e = np.empty((a.size+1,), dtype=a.dtype)
            half_da = (a[1:] - a[:-1])/2
            e[0] = a[0] - half_da[0]
            e[1:-1] = a[:-1]+half_da
            e[-1] = a[-1] + half_da[-1]
            return e
        
        p_edges = np.exp(get_edges(np.log(self.press)))
        w_edges = get_edges(wave_points.value)
        
        fig.suptitle(f'Opacities for {RadtranGasDescriptor(self.gas_id, self.iso_id).label} at each point on the temperature grid.')
        
        for i in range(self.k.shape[2]):
            ax[i].pcolormesh(
                w_edges,
                p_edges,
                #self.k[:,:,i].T
                np.log(self.k[:,:,i].T)
            )
            
            if wave_points.unit == ans.enums.WaveUnit.Wavenumber_cm:
                ax[i].set_xlabel('Wavenumber ($cm^{-1}$)')
            elif wave_points.unit == ans.enums.WaveUnit.Wavelength_um:
                ax[i].set_xlabel(r'Wavelength ($\mu$m)')
            else:
                raise RuntimeError(f'Unknown {wave_unit}. Should be one of {ans.enums.WaveUnit.values()}')
            
            ax[i].set_xlabel('Wavenumber ($cm^{-1}$)')
            ax[i].set_ylabel('Pressure (atm)')
            ax[i].set_title(f'Temperature: {self.temp[i]} (K)')
        
        plt.show()



def read_legacy_header(f : str | io.IOBase) -> LblHeader:
    if not isinstance(f, io.IOBase):
        with open(f, 'rb') as g:
            return read_legacy_header(g)
    
    buf = f.read(_header_struct.size)    
    return LblHeader(*_header_struct.unpack_from(buf))


def read_legacy(
        f : str | io.ioBase, 
        wave_unit : None | ans.enums.WaveUnit =  None,
        fortran_record_byte_size = 4, # number of bytes in a fortran record, normally 4
) -> LblDataTProfilesAtPressure | LblDataTPGrid:
    """
    Read line-by-line table written in legacy format. NOTE: factor of 1E-20 (reverse of 
    factor applied when storing to avoid underflow in single precision floats) IS NOT 
    applied here, it is applied later in ForwardModel_0.
    
    ## ARGUMENTS ##
        
        f : str | ioBase
            Filename or FileLike object for table to read.
        
        wave_unit : None | ans.enums.WaveUnit
            Unit of wave data in file,. If None will infer from value of `LblHeader.vmin`; < 100 -> Wavelength_um, >= 100 -> Wavenumber_cm.
    
    ## RETURNS ##
    
        lbl_data : LblDataTProfilesAtPressure | LblDataTPGrid
            Line-by-line data in one of the two specified formats depending upon the 
            format specified in the file.
    """
    if not isinstance(f, io.IOBase):
        with open(f, 'rb') as g:
            return read_legacy(g, wave_unit)
    
    _lgr.debug(f'{f.tell()=}')
    hdr = read_legacy_header(f)
    _lgr.debug(f'{hdr=}')
    
    if wave_unit is None:
        if hdr.vmin < 100:
            wave_unit = ans.enums.WaveUnit.Wavelength_um
        else:
            wave_unit = ans.enums.WaveUnit.Wavenumber_um
        _lgr.warning(f'No `wave_unit` parameter given to {__name__}.read_legacy(...). Assuming "{wave_unit.name}" based off `LblHeader.vmin` = {hdr.vmin}')
    
    ptk = [
        None,
        None,
        None
    ]
    
    
    
    fortran_record_shape, p_struct, t_struct, k_struct = _get_body_structs(hdr.nwave, hdr.npress, hdr.ntemp)
    
    p_shape = (hdr.npress,)
    if hdr.ntemp < 0:
        t_shape = (hdr.npress, -hdr.ntemp)
        k_shape = (hdr.nwave, hdr.npress, -hdr.ntemp)
        lbl_data_type = LblDataTProfilesAtPressure
    else:
        t_shape = (hdr.ntemp,)
        k_shape = (hdr.nwave, hdr.npress, hdr.ntemp)
        lbl_data_type = LblDataTPGrid
    
    _lgr.debug(f'{p_shape=} {t_shape=} {k_shape=} {lbl_data_type=}')
    _lgr.debug(f'{f.tell()=}')
    
    # Read in pressure and temperature data, these start directly after the header
    for i, (x_struct, x_shape) in enumerate([(p_struct, p_shape), (t_struct, t_shape)]):
        _lgr.debug(f'{i=} {x_struct=} {x_shape=}')
        buf = f.read(x_struct.size)
        _lgr.debug(f'{f.tell()=}')
        
        ptk[i] = np.array(x_struct.unpack_from(buf), dtype=float, order='F').reshape(x_shape)
    
    # Absorption coefficients start at `hdr.irec0` records into the file. THIS IS NOT THE SAME AS AFTER THE PRESSURE AND TEMPERATURE DATA!!
    current_pos = f.tell()
    n_bytes_from_start_of_file_to_abs_coeff_data = fortran_record_byte_size * (hdr.irec0-1)
    
    if current_pos > n_bytes_from_start_of_file_to_abs_coeff_data:
        _lgr.warning(f'Header.IREC0 {hdr.irec0} is not correct for this file, attempting to guess the correct location for absorption coefficient data')
        f.seek(-k_struct.size, 2) # Assume the coefficients end and the end of the file and go backwards to where they should start
    else:
        _lgr.debug(f'Skipping to Header.IREC0 {hdr.irec0} bytes into the file as that is where the absorption coefficient data should be. {n_bytes_from_start_of_file_to_abs_coeff_data=}')
        f.seek(n_bytes_from_start_of_file_to_abs_coeff_data, 0)
    
    assert ptk[2] is None, 'Nothing should have been stored here before this point'
    ptk[2] = np.array(k_struct.unpack_from(f.read(k_struct.size)), dtype=float).reshape(k_shape)
    
    # Check we have read all the data correctly
    current_pos = f.tell()
    f.seek(0, 2)
    end_pos = f.tell()
    
    if current_pos != end_pos:
        _lgr.warning(f'We have not read all the data from the file, there are {end_pos - current_pos} bytes remaining. Using an alternate method to try and ensure we get the data correctly.')
        f.seek(-k_struct.size, 2) # Assume the coefficients end and the end of the file and go backwards to where they should start
        ptk[2] = np.array(k_struct.unpack_from(f.read(k_struct.size)), dtype=float).reshape(k_shape)
    
    return lbl_data_type(hdr.gas_id, hdr.iso_id, wave_unit, np.linspace(hdr.vmin, hdr.vmin+hdr.delv*hdr.nwave, hdr.nwave, endpoint=False), *ptk)
    
    








